/* This is bruhat.c

  Coxeter version 1.0  Copyright (C) 1991 Fokko du Cloux
  See file main.c for full copyright notice
*/

#include "bruhat.h"

/*
	The main objective of this module is to construct the Bruhat order. 
	In this version, the approach is totally "brute force" (although pretty 
	efficient as far as speed is concerned). Indeed the Bruhat order for the 
	whole group is constructed in one huge bit-matrix by the function 
	MakeBrOrder(). When the program is ran for E6, it never gets past here.
	It is hoped that this will be changed in the near future to a more
	modular approach, where the pieces of the Bruhat order for each length
	are constructed separately, and so may be forgotten when no longer needed.
	
	The idea is to first get hold of the "principal Bruhat cells" for 
	each w, i.e. the cells in codimension 1. These are stored in the list
	pointed by prbruhat, consisting of a row for each w. The Bruhat order
	is then constructed recursively, by remarking that the Schubert cell
	of w is the union of w and the Schubert cells of the elements in 
	prbruhat[w].
	
        When dealing with Bruhat order questions, the best Weyl group
	representation is usually through weylnbr_t, i.e. w is represented by its
	number in the lexicographical order on normal forms. In order to pass
	easily from weylnbr_t to weylarr_t, we construct transition tables :
	lexorder+rank*w points to a weylarr_t which is the array-form of element 
	no. w in lexicographical order; and natorder+j points to a weylnbr_t
	which corresponds to the array no.j in the "natural" (inverse) 
	lexicographical order on arrays. Whereas these tables may seem wasteful 
	in space, actually their size is rather negligible with respect to that 
	of brorder; so it doesn't really make much difference (the reason we 
	did not use the "natural" lexicographical order throughout is that it 
	is not compatible with the Bruhat order; not even with the lengths).
	
	We also store as a table the elementary multiplications in W in the 
	weylnbr_t representation; this amounts to constructing an automaton based 
	on the whole group instead of the sequence of small automata that we used 
	so far (actually, we store the two-sided transition tables). Although this
	is not very satisfactory, it appears that even if this big automaton is 
	what is eventually needed, having the small ones at our disposal makes for 
	a huge increase in speed in the construction of the big one. It should be 
	pointed out that this construction is not really necessary, but it 
	improves speed a lot, so as long as there is enough memory, there seems 
	to be no reason to do without it. This transition table is lexprod.
*/

PreBruhat()

 /*

   Constructs all the preliminaries to the Bruhat order (we put these aside
   in case they are needed independently)
   
 */

{

InitConstants();

MakeNLOrders();

MakeInverse();

MakeLexProd();

MakeDownFlags();

MakePrBruhat();

}   /* end of PreBruhat */

Bruhat()

/**********************************************************************

	Constructs brorder, and also natorder, lexorder and inverse, since
these are needed along the way. After calling Bruhat, all the strings
appearing as global variables are initialized and get their proper values.

***********************************************************************/

{

PreBruhat();

InitBruhat();	/* allocate memory */

MakeBrOrder();

}	/* end of Bruhat */


MakeBrOrder()

/*
  Makes the Bruhat order using MakeBrPiece. This function does exactly
  the same thing as the old MakeBrOrder.
*/

{
  int p;
  lflags_t **old,**new;

  **brorder = wordmask[0];

  new = brorder;

  for (p=1;p<=maxlength;p++)
    {
      old = new;
      new += betti[p-1];
      MakeBrPiece(p,old,new);
    }
}    /* end of MakeBrOrder_new */


MakeBrPiece(p,old,new)

/*
  
  This function returns in new the part of the Bruhat-order bit-matrix 
  corresponding to elements of length p. It is assumed that old points
  to the previous part, i.e. that corresponding to length p-1. Together with
  the knowledge of the prbruhat array, this is all that we need to know. It
  is assumed that new has been properly allocated -- in other words,
  that it points to betti[p] rows of rowlength[p] lflags_t, all set to 0.

*/

     lengthcount_t p;
     lflags_t **old, **new;

{
  weylnbr_t i,*a;
  weylsize_t b,b1;
  unsigned r,r1,k,k1;
  lflags_t f;
  lengthcount_t j;
  

  if (p==0)
    {
      **new |= wordmask[0];
      return(0);
    }

  r = rowlength[p-1];

  b1 = 0;
  for (j=0;j<p-1;j++)
    {
     b1 += betti[j] ;
    }
  /* b1 is the origin of old */

  b = b1 + betti[p-1];
  /* b is the origin of new */

  for (i=1;i<betti[p];i++)    /* make row pointers */
    {
      new[i] = new[i-1] + rowlength[p];
    }

  for (i=0;i<betti[p];i++)
    {
      new[i][(b+i)/wordsize] = wordmask[(b+i)%wordsize];
      
      r1 = prbruhat[b+i].size;

      for (k=0;k<r;k++)
	{
	  a = (weylnbr_t *)prbruhat[b+i].row;
	  f = 0;
	  for (k1=0;k1<r1;k1++)
	    {
	      f |= old[*a-b1][k];
	      a++;
	    }
	  new[i][k] |= f;
	}
    }

}    /* end of MakeBrPiece */


InitBruhat()

/**********************************************************************

	Allocates the memory needed for brorder and writes row adresses.

***********************************************************************/

{
weylnbr_t i,j;
int p;

CALLOC(brorder,wsize,lflags_t *,
	printf(" for brorder\nwsize = %lu",(unsigned long)wsize);)
CALLOC(*brorder,brsize,lflags_t,
	printf("\nbrsize = %lu",(unsigned long)brsize);)

for (p=0,j=0;p<maxlength;p++)
	{
	for (i=0;i<betti[p];i++,j++)
		{
		brorder[j+1] = brorder[j] + rowlength[p];
		}
	}
}	/* end of InitBruhat */


FreeBruhat()

/**********************************************************************

	This function frees the memory allocated by InitBruhat

***********************************************************************/

{
;
}	/* end of FreeBruhat */


static InitConstants()

/**********************************************************************

Initializes the following constants :

	- charmask : charmask[i] gets bit #i in an unsigned char (starting on
	  the right, and starting with i=1.)
	  
	- dfmask : same as charmask, for type downflags_t
	  
	- wordmask : same for an unsigned long int (starting from zero now.)
	
	- rightmask and leftmask (for taking the right or left flags in 
	  downflags)
	
	- wordsize : size of a long int, in bits
	
	- firstlength : accumulates betti (so firstlength[p] will hold the
	  adress in lexorder of the first element of length p)
	  
	- rowlength : length of rows in the brorder matrix.
	
	- brsize : size of the bruhat matrix, in long ints.
	
	- lexprodstring : workstring for LexRProd

***********************************************************************/

{

int i,j,l;
weylsize_t c;


CALLOC(charmask,rank+1,unsigned char,;)
CALLOC(dfmask,CHARBITS*sizeof(downflags_t)+1,downflags_t,;)
CALLOC(wordmask,CHARBITS*sizeof(lflags_t),lflags_t,;)

charmask[1] = 1;
for (l=2;l<=rank;l++)
	{
	charmask[l] = charmask[l-1] << 1;
	}
dfmask[1] = 1;
for (i=2;i<=CHARBITS*sizeof(downflags_t);i++)
	{
	dfmask[i] = dfmask[i-1] << 1;
	}
wordmask[0] = 1;
for (i=1;i<CHARBITS*sizeof(lflags_t);i++)
	{
	wordmask[i] = wordmask[i-1] << 1;
	}

rightmask = 1;
rightmask <<= rank;
rightmask -= 1;	/* it is assumed that downflags_t holds at least 2*rank bits */

leftmask = rightmask;
leftmask <<= rank;

wordsize = CHARBITS*sizeof(long int);
CALLOC(rowlength,maxlength+1,unsigned short,;)

for (j=0,c=0,brsize=0;j<=maxlength;j++)
	{
	  c += betti[j];
	  if (c%wordsize)
	    {
	      rowlength[j] = c/wordsize + 1;
	    }
	  else
	    {
	      rowlength[j] = c/wordsize;
	    }
	  brsize += betti[j]*rowlength[j];
	}

CALLOC(lexprodstring,rank,weylpiece_t,;)

}	/* end of InitConstants */


MakeNLOrders()

/***********************************************************************

	This function allocates the pointers lexorder and natorder. It returns
	in lexorder a list of the array-forms of elements of W, in the lexico-
	graphical ordering of their normal forms; and in natorder, a list of
	the lex-order numbers of the elements of W, written in the "natural
	order" of their array forms (i.e. inverse lexicographical order of
	the arrays, actually). The two arrays are constructed together for 
	efficiency (otherwise natorder would be rather nasty to get). It is 
	assumed that the vector subwsizes has been initialized.

************************************************************************/

{
weylarr_t w,ws,firstfree;
generator_t s;
weylnbr_t j;
register weylnbr_t a;
int i,l;

/* memory allocation */

CALLOC(lexorder,rank*(unsigned long)wsize,weylpiece_t,
	printf(" for lexorder\nrequest = %lu bytes",rank*(unsigned long)wsize);)
CALLOC(natorder,wsize,weylnbr_t,
	printf("for natorder\nwsize = %lu",(unsigned long)wsize);)
CALLOC(ws,rank,weylpiece_t,;)

/* generation of lxorder and ntorder */

j = 1;

for (w=lexorder+rank,l=1;l<=rank;w+=rank,l++)
	{
	*(w+l-1) = 1;
	*(natorder+wsubsize[l-1]) = j++;
	}
firstfree = lexorder + rank*(rank+1);

for (w=lexorder+rank;w<firstfree;w+=rank)
	{
	for (s=1;s<=rank;s++)
		{
		Prod(w,s,ws);
		if (Lastterm(ws)==s)	/* ws is normal */
			{
			for (a=0,i=0;i<rank;i++)
				{
				*(firstfree++) = ws[i];
				a += ws[i]*wsubsize[i];
				}
			natorder[a] = j++;
			}
		}
	}

}	/* end of MakeNLOrders */


weylnbr_t LexOrder(w)

/*********************************************************************

	Returns the lexicographic order of an element of the Weyl group
	represented as an array. Uses the array natorder.

**********************************************************************/

	weylarr_t w;

{
weylnbr_t a=0;
int j;

for (j=0;j<rank;j++)
	a += w[j]*wsubsize[j];

return(natorder[a]);

}	/* end of LexOrder */


MakeDownFlags()

/************************************************************************

	This function allocates and constructs the downflags. The n-th bit
	in the left (resp. right) field of a structure DownFlags is set if
	multiplication by generator n+1 on the left (resp. right) takes the
	length of the corresponding Weyl group element down. We assume
	MakeLexOrder has already been called.
	
	NOTE : This function could probably be speeded up quite a bit !

*************************************************************************/

{
downflags_t *dflags;
weylnbr_t j;
weylarr_t w,x;
state_t *state,**state0;
generator_t s,t;
int l;
	
CALLOC(downflags,wsize,downflags_t,
	printf(" for downflags\nwsize = %lu",(unsigned long)wsize);)

/*
cutdata_t *cdata;

for (w=1;w<wsize;w++)
	{
	cdata = CutOff(w);
	
	}
*/

dflags = downflags + wsize;

for (w=lexorder+rank*wsize;w>lexorder;w-=rank) /* set right downflags */
	{
	dflags--;
	for (s=1;s<=rank;s++)
		{
		/* check if length of ws goes down */

		for (t=s,l=rank,state0=table+rank-1,x=w-1;;state0--,l--,x--)
			{
			state = *state0 + (l+1)**x;
			if (*(state+t) >= 0)
				{
				if (*(state+t) <= *x)
					*dflags |= dfmask[s];
				break;
				}
			else
				t = -*(state+t);
			}
		}
	}

/* set left flags by looking at the inverses */

{
register downflags_t f;

for (j=0;j<wsize;j++)
	{
	f = downflags[inverse[j]] & rightmask;
	f <<= rank;
	downflags[j] |= f;
	}
}

}	/* end of MakeDownFlags */


MakeInverse()

/********************************************************************

	This function allocates and returns a pointer to the table of
	inverses of the Weyl group elements, written in the lexicographical
	order. This is used, for example, to pass from the right downflags
	to the left ones. It assumes MakeLexOrder has been called.
	
	NOTE : could be speeded up.

*********************************************************************/

{
weylnbr_t a,j;
weylarr_t w,w1;
int i;

CALLOC(inverse,wsize,weylnbr_t,
	printf("for inverse\nwsize =%lu",(unsigned long)wsize);)

CALLOC(w1,rank,weylpiece_t,;)

for (j=0,w=lexorder;j<wsize;j++,w+=rank)
	{
	Inverse(w,w1);
	for (a=0,i=0;i<rank;i++)
		a += wsubsize[i]*w1[i];
	inverse[j] = natorder[a];
	}
	
free(w1);

}	/* end of MakeInverse */


MakePrBruhat()

/********************************************************************

	Each row of prbruhat contains the "principal" Bruhat cells for
the corresponding Weyl group element, i.e. the cells in codimension 1.
The rows are constructed by induction on the length, by removing the 
last element in the normal form of w, say s, and looking at prbruhat[ws].
It is enough to see which elements in prbruhat[ws] remain reduced after 
right multiplication by s, and of course add ws itself to the list thus 
obtained. By the way, an interesting conjecture is that there should be 
at most 2*rank elements in each prbruhat[w] (it is known that there are 
at least rank such elements if w is of full support).

*********************************************************************/

{
weylnbr_t *a,*a0;
weylnbr_t i,j,k;
int p;
struct CutData cdata;
zoneinfo_t *zone;

CALLOC(a0,maxlength + 1,weylnbr_t,;)

zone = AllocZone(sizeof(weylnbr_t));

CALLOC(prbruhat,wsize,rowhead_t,;)

prbruhat[0].row = zone->firstfree;
prbruhat[0].size = 0;

for (j=1;j<wsize;j++)	/* now we are in length >= 3 */
	{
	a=a0;
	cdata = CutOff(j);	/* cuts off the last term */
	i = cdata.i;
	
	for (p=0;p<prbruhat[i].size;p++)
		{
		k = ((weylnbr_t *)(prbruhat[i].row))[p];
		if(downflags[k] & dfmask[cdata.s])
			;	/* length goes down */
		else
			{
			LexRProd(k,cdata.s,a++);
			}
		}
	*(a++) = i;	/* i is always in prbruhat[j] */
	
	/* write a0 in prbruhat[j] */
	
	p = a-a0;
	
	if (p > zone->freestore)	/* get more storage */
		{
		NewBlock(zone,sizeof(weylnbr_t));
		}
	
	prbruhat[j].row = zone->firstfree;
	prbruhat[j].size = p;
	zone->firstfree += (p)*sizeof(weylnbr_t);
	zone->freestore -= p;
	
	{ 
	register int i;
	register weylnbr_t *a,*b;
	
	a = (weylnbr_t *)(prbruhat[j].row);
	for (i=0,b=a0;i<p;i++)
		{
		*(a++) = *(b++);
		}
	}
	}

free(a0);
return;

}	/* end of MakePrBruhat */


struct CutData CutOff(i)

	weylnbr_t i;

/********************************************************************

	This is an auxiliary to MakePrBruhat. It returns the last term
	of an element written in lexorder, and the result, also in lexorder,
	of cutting off this last term. The idea is not to do the same work
	twice, by first calling lastterm, and then prod.

*********************************************************************/

{
weylarr_t w;
state_t *state;
int j = rank-1;
struct CutData cdata;
weylnbr_t a;

w = lexorder + i*rank;
while (w[j] == 0)
	j--;

if (j < 0)
	{
	cdata.s = 0;
	return(cdata);
	}
else
	{
	state = table[j]+(j+2)*w[j];
	cdata.s = *state;
	a = wsubsize[j--]**(state+*state);
	while (j >= 0)
		{
		a += w[j]*wsubsize[j];
		j--;
		}
	cdata.i = natorder[a];
	return(cdata);
	}

}	/* end of CutOff */


LexRProd(w,s,ws)

/**********************************************************************

	Returns in ws the product of w -- a WEYLNBR, not a 	WEYLARR as in
Prod; returns 1 if the length goes up, 0 if it goes down.

***********************************************************************/

		weylnbr_t w,*ws;
		int s;

{
weylarr_t w1;
int retvalue;

w1 = lexorder + rank*w;
retvalue = Prod(w1,s,lexprodstring);
*ws = LexOrder(lexprodstring);
return(retvalue);

}	/* end of LexRProd */


LexLProd(w,s,sw)

/***********************************************************************

	Returns in sw the product of s and w; similar to LexRProd.

************************************************************************/

	weylnbr_t w,*sw;
	int s;

{
int retvalue;

w = inverse[w];
retvalue = LexRProd(w,s,sw);
*sw = inverse[*sw];
return(retvalue);

}	/* end of LexLProd */


MakeLexProd()

/************************************************************************

	Puts in lexprod[w*2*rank+j] the weylnbr corresponding to ws, s=j+1, if j<rank,
to sw, s=j-rank+1 if rank<=j<2*rank; this is for fast computation of such
products later on.

*************************************************************************/

{

weylnbr_t w;
register int j;

/* memory allocation */

CALLOC(lexprod,2*rank*(unsigned long)wsize,weylnbr_t,
	printf(" for lexprod\nrequest = %lu bytes",
	2*rank*(unsigned long)wsize);)

for (w=0;w<wsize;w++)
	{
	for (j=0;j<rank;j++)
		{
		LexRProd(w,j+1,lexprod+2*w*rank+j);
		}
	for (j=0;j<rank;j++)
		{
		LexLProd(w,j+1,lexprod+(2*w+1)*rank+j);
		}
	}

}	/* end of MakeLexProd */
