/*
  This is types.h

  Coxeter version 1.0  Copyright (C) 1991 Fokko du Cloux
  See file main.c for full copyright notice
*/

                          /* Interface */

struct CommandCell
  {
    int command;
    char letter;
    struct CommandCell *left;
    struct CommandCell *right;
  };

typedef struct CommandCell ccell_t;


			  /* Memory management */


struct ZoneInfo
	{
	char **first;
	char **last;
	char *firstfree;		 /* to be typecast */
	unsigned freestore;		 /* in units of the type stored */
	};

typedef struct ZoneInfo zoneinfo_t;

typedef unsigned rowsize_t;             /* size of a row */
typedef unsigned long lrowsize_t;       /* for large memory allocations */

struct RowHead
	{
	char *row;			/* to be typecast */
	rowsize_t size;		        /* in units of type stored */
	};

typedef struct RowHead rowhead_t;


/*	The primitive types, in terms of which everything else is defined, 
	are the following :
	
	- typeid_t for the type of the Weyl group (A,B,C,D,E,F,G,H for the 
	  irreducible finite Coxeter groups) : char in this implementation;
	  
	- rankid_t for the rank : int in this implementation;
	
	- generator_t for the generators : int in this implementation;
	
	- weylletter_t for the generators when they are part of a string : 
	  unsigned char in this implementation;
	  
	- weylpiece_t for the distinguished elements in the subquotients :
	  unsigned short in this implementation;
	  
	- weylnbr_t for the lexicographical-order representation of the Weyl
	  group elements : unsigned in this implementation;
	  
	- state_t for the entries in the transition tables of the automata :
	  int in this implementation;
	  
	- weylsize_t for sizes of objects not larger than the group : unsigned 
	  long in this implementation;
	  
	- quotsize_t for objects not larger than the largest subquotient : 
	  unsigned long in this implementation;
	  
	- lengthcount_t for objects not larger than maxlength (plus some 
	  overhead perhaps) : unsigned short in this implementation;
	  
	- lflags_t for large chunks of bit-arrays (it is more efficient to
	  cut them in rather large chunks, even if this costs some space) :
	  unsigned long in this implementation;
	  
	- sflags_t for smaller chunks : unsigned short in this implementation;
	
	- downflags_t for flagging the generators that take a given weyl group 
	  element down on right or left multiplication : unsigned short in this 
	  implementation;
	  
	- klcoeff_t for the coefficients of Kazhdan-Lusztig polynomials : 
	  short in this implementation.
	  
		Of course some compatibility has to be maintained among these : a 
	weylletter_t must fit in a generator_t (the reverse direction is the 
	programmer's responsibility, just as the game between char's and int's 
	for C strings); a weylpiece_t must fit in a state_t (actually we have 
	not been fully rigorous here; in fact quantities of type weylpiece_t 
	should not exceed half their range, so that they remain positive after 
	conversion to signed quantities); a weylpiece_t should fit in a 
	quotsize_t; a weylnbr_t should fit in a weylsize_t. Also downflags_t 
	should hold at least 2*rank bits.
	
		Provided the compatibilities above are maintained, a change in
	one of the primitive types, to accomodate the available hardware for
	example, or to treat a specific problem for which the choices above
	are unappropriate, should prove harmless. */


				/* Types for Weyl group routines */
						

/* types for loop indices and counts */

	typedef unsigned long weylsize_t;	/* primitive type */
	typedef weylsize_t quotsize_t;		/* primitive type */
	typedef unsigned short lengthcount_t;	/* primitive type */

/* Types for statistic data */

/* 	This might get more complicated in the future if non-simple groups are 
	required */

typedef char typeid_t;			/* primitive type */
typedef int rankid_t;			/* primitive type */

struct InitData
	{
	  quotsize_t quotsize;
	  lengthcount_t maxqlength;
	  typeid_t subtype;
	};

typedef struct InitData initdata_t;

struct WeylData
	{
	  quotsize_t quotsize;
	  lengthcount_t maxlength;
	  lengthcount_t maxqlength;
	};

typedef struct WeylData weyldata_t;

/* types for the string representation of Weyl group elements */

	typedef unsigned char weylletter_t;
	typedef weylletter_t *weylword_t;

/* type for the generators */

	typedef int generator_t;			/* primitive type */

/* type for the entries in the automaton-tables */

	typedef int state_t;				/* primitive type */

/* types for the array representation of Weyl group elements */

/* weylpiece_t should be compatible with state_t */

	typedef unsigned short weylpiece_t;		/* primitive type */
	typedef weylpiece_t *weylarr_t;

/*
  type for the number representation (e.g. in lexorder) of Weyl group
  elements. Note that weylnbr_t may be smaller than weylsize_t (this
  happens for example on the Macintosh). The reason for the two types
  is that in memory-critical or speed-critical situations, it seems
  to be worthwile to use types as small as possible. This is obvious
  for memory, but it is also very noticeable for speed. Anyway, the
  short types are (barely) enough for rank up to 6, so they cover all
  cases for which kl-computations are currently feasible.
*/

	typedef unsigned weylnbr_t;                    /* primitive type */

	
       			/* Types for Bruhat order routines */
	

/* types for bit-matrices */

	typedef unsigned long lflags_t;			/* primitive type */
	typedef unsigned short sflags_t;		/* primitive type */

/* types for Bruhat */
	
struct CutData{
	weylnbr_t i;
	generator_t s;};

typedef unsigned short downflags_t;			/* primitive type */
/* suffices for rank up to 8 */


				/* Types for Kazhdan-Lusztig routines */


/* types for Kazhdan-Lusztig polynomials */

	typedef short klcoeff_t;			  /* primitive type */
	typedef klcoeff_t *klpol_t;
        typedef weylsize_t *bettipol_t; /* type for betti-number polynomials */
        typedef short rcoeff_t;
        typedef rcoeff_t * rpol_t;

/* in this version we allocate room for the computation of the R-polynomials
   as well */

struct kldata		/* type for the cells in kl_list */
	{
	weylnbr_t y;
	weylnbr_t ymin;
	klpol_t *pol;
	rpol_t rpol;    /* for the R-polynomial */
	};

typedef struct kldata kldata_t;
	
struct klcell
	{
	klpol_t pol;
	struct klcell *left;
	struct klcell *right;
	};

typedef struct klcell klcell_t;

struct betticell
       {
	 bettipol_t pol;
	 struct betticell *left;
	 struct betticell *right;
       };

typedef struct betticell betticell_t;

struct MuData
	{
	weylnbr_t y;
	unsigned short height;	/* holds (l(w)-l(y)-1)/2 = deg(P_{y,w}) */
	klcoeff_t mu;
	};

typedef struct MuData mudata_t;


/* type for monitoring the state of the program */

struct Status
  {
    typeid_t type;
    rankid_t rank;
    int size;
    int representation;
    int mode;
    int weyldone;
    int bruhatdone;
    int kldone;
  };

typedef struct Status status_t;
