/*
  This is weyl.c

  Coxeter version 1.0  Copyright (C) 1991 Fokko du Cloux
  See file main.c for full copyright notice
*/

#include "weyl.h"

/*
		We use three different representations of Weyl group elements.
		
		First of all, we have the represention of w \in W as a string 
	of generators; this is the type weylword_t, the generators being of 
	type weylletter_t (unsigned char in this implementation). 
	
		Second,we can represent w as a product of rank distinguished 
	elements, eachin one of the subquotients of the tower of subgroups 
	defined by the chosen order on the generators; this is the type 
	weylarr_t, each component of the array being of type weylpiece_t 
	(unsigned short in this implementation -- for a reasonable ordering 
	of the generators, this is more than enough even for type E8).
	
		Third, we can represent w as a single number; this amounts to 
	choosing an ordering of the group. We have chosen the lexicographical 
	ordering on the normal forms, first ordering by length -- this is
	not so nice to work with, but it has the advantage of being compatible
	with Bruhat order. This is the type weylnbr_t (unsigned in this 
	implementation -- this may turn out to be rather small, in fact it is
	the same as unsigned short on the Mac, but even this is enough for E6,
	and for groups for which this is too small it is usually out of the
	question to carry out any computation involving Bruhat order, which is 
	the only situation where this type is needed). Anyway it is easy enough 
	to modify this (see the comments in types.h).
*/


Weyl()

/*********************************************************************

	Does the necesary allocations and initializations to be able to use
	all the Weyl group functions. It only assumes rank and type have
	been assigned the right values; these have to remain constant
	throughout.

**********************************************************************/

{

InitAutomaton();
MakeAutomaton();
InitConstants();

}	/* end of Weyl */


static InitConstants()

/*********************************************************************

Allocates and initializes workspace for Inverse, Mult, NormalForm,
StringMult, Power. Initializes the following constants :

	- the array wsubsize (sizes of the subgroups in the filtration);
	
	- wsize (order of the Weyl group);
	
	- maxlength (length of longest element);

	- longarray (array holding the longest element);
	
	- the array betti (betti numbers of the flag variety, or 
	  equivalently distribution of the groupe elements according to 
	  length);
	  
	- the array firstlength; firstlength[p] is the first element of
	  length p; for convenience firstlength[maxlength+1] is defined to 
	  be wsize;

	- the array star; star[s-1] is the support of the "star" of s
	  in S, i.e. it flags those t for which Cox(s,t) > 2;

	- the array bettiwork; it serves as a workstring for the
	  function BettiPol (which is in kl.c). It is of length
	  maxlength+2 (maxlength+1 coefficients, plus one for the
	  degree);

	- the array multwork; it serves as a workstring for the
	  function BPolMult.

	- the arrays xlength and xlist; they are data for the function
	  WeylnbrToArray

**********************************************************************/

{
  rankid_t l;
  lengthcount_t j;
  quotsize_t count;

  maxlength = head[rank-1]->maxlength;

  CALLOC(invstring,maxlength+2,weylletter_t,;)
  CALLOC(multstring,maxlength+1,weylletter_t,;)
  CALLOC(bstring,maxlength+1,weylsize_t,;)
  CALLOC(print_string,maxlength+1,weylletter_t,;)
  CALLOC(nfarray,rank,weylpiece_t,;)
  CALLOC(fgarray,rank,weylpiece_t,;)
  CALLOC(power_arr,rank,weylpiece_t,;)
    
  CALLOC(bpol1_work,maxlength+2,weylsize_t,;)
  CALLOC(bpol2_work,maxlength+2,weylsize_t,;)

  CALLOC(multwork,maxlength+2,weylsize_t,;)
  CALLOC(divwork,maxlength+2,weylsize_t,;)

  CALLOC(wsubsize,rank,weylsize_t,;)
    
  *wsubsize = 1;
  for (l=1;l<rank;l++)
    *(wsubsize+l) = *(wsubsize+l-1)*head[l-1]->quotsize;

  wsize = wsubsize[rank-1]*(**(head+rank-1)).quotsize;

  CALLOC(longarray,rank,weylpiece_t,;)

  for (l=0; l<rank; l++)
    {
      longarray[l] = (weylpiece_t)(head[l]->quotsize - 1);
    }

  CALLOC(betti, maxlength+1, weylsize_t, ;)

  /* betti is an array; the elements in qbetti are polynomials (i.e. their
     first term is the degree). qbetti[l] holds the betti numbers of the
     quotient of W by W_{l-1} (so that qbetti[1] is the same as betti,
     except for the degree) */

  CALLOC(qbetti, rank, bettipol_t, ;)
  j = 0;
  for (l=rank; l>0; l--)
    {
      j += head[l-1]->maxqlength;
      CALLOC(qbetti[l-1], j+2, weylsize_t, ;);
    }
  
  MakeBetti();

  CALLOC(firstlength,head[rank-1]->maxlength+2,weylsize_t,;)
  *firstlength = 0;
  for (j=0;j<=maxlength;j++)
    {
      firstlength[j+1]=firstlength[j]+betti[j];
    }

  CALLOC(star,rank,sflags_t,;)
  MakeStar();

  CALLOC(xlength,rank,lengthcount_t *,;)
  CALLOC(xlist,rank,weylpiece_t *,;)

  for (l=0;l<rank;l++)
    {
      CALLOC(xlength[l],head[l]->quotsize,lengthcount_t,;)
      CALLOC(xlist[l],head[l]->quotsize,weylpiece_t,;)
      count = 0;
      MakeXLength(0,l,0,&count);
      count = 0;
      MakeXList(0,l,&count);
    }

    }				/* end of InitConstants */


static MakeStar()

/*
  Fills in the array star
*/

{
  register sflags_t f;
  generator_t s,t;
      
  for(s=1;s<=rank;s++)
    {
      t = rank;
      f=0;
      
      do
	{
	  if (Cox(s,t)>2)
	    {
	      f++;
	    }
	  if (t-- == 1)
	    break;
	  f <<= 1;
	}
      while (1);

      star[s-1] = f;
    }
  
}    /* end of MakeStar */


static MakeXLength(x,l,p,count)

/*
  Recursively fills in the data in xlength[l]
*/

     lengthcount_t p;
     weylpiece_t x;
     quotsize_t *count;
     rankid_t l;

{

  state_t *state, *state1;
  weylpiece_t x1;
  rankid_t j;

  xlength[l][*count] = p;  /* length of current piece */
  *count += 1;

  state = table[l] + x*(l+2);

  for (j=l+1;j>0;j--)
    {
      if (state[j] < (state_t)x)
	continue;

      x1 = state[j];
      state1 = table[l] + x1*(l+2);
	  
      if(*state1 == j)		/* j is the last term of x1 */
	{
	  MakeXLength(x1,l,p+1,count);
	}
    }

}    /* end of MakeXLength */


static MakeXList(x,l,count)

/*
  Recursively fills in the data in xlist[l]
*/

     weylpiece_t x;
     quotsize_t *count;
     rankid_t l;

{

  state_t *state, *state1;
  weylpiece_t x1;
  rankid_t j;

  xlist[l][*count] = x;  /* current piece */
  *count += 1;

  state = table[l] + x*(l+2);

  for (j=l+1;j>0;j--)
    {
      if (state[j] < (state_t)x)
	continue;

      x1 = state[j];
      state1 = table[l] + x1*(l+2);
	  
      if(*state1 == j)		/* j is the last term of x1 */
	{
	  MakeXList(x1,l,count);
	}
    }

}    /* end of MakeXList */


MakeAutomaton()

/********************************************************************

	This function constructs the transition tables of the automaton
previously initialized by InitAutomaton (see the description
of the memory zone there). Here we will describe those tables in
more detail. For each of the rank state_t-pointers starting at
*table, there will be quotsize(l) cells,
each of size (l+1)*sizeof(state_t), where l is the current rank.

	The first field holds the last term of the normal form of the
shortest coset representative, thus allowing to reconstruct this
normal form inductively -- this is essentially what the function 
NbrToStringpiece does. The next l fields hold either the address of
xs if xs is a coset representative, or -t if xs = tx, in which
case we remain in the same coset (it is known that one of these two
cases must hold). In the comments X is the set of shortest coset
representatives.

	The tables are filled in by induction on the length (actually,
the elements are written in lexicographical order on the shortest
representatives). The process is organized carefully in order that
whenever a function like NumberToString is called, which uses the
transition tables, only the part of the tables that is already
constructed is needed. The fact that this is so is not obvious, and
needs sometimes a nontrivial proof. Also, some of the if-else
branchpoints use theoretical results to get rid of certain
possibilities. For a description of the normal-form algorithm, see
the documentation. In the process of constructing the tables, the
lengthcounts are aso filled in.

********************************************************************/

{

struct WeylData *wdata;
rankid_t l;
int j,p,firstfree;
state_t *state,*state0,*newcell,*cell;
generator_t s,t;
quotsize_t thislength,newlength=0,*lc;
weylword_t f,f1;
weylarr_t v,v1,v2;
weylpiece_t x,xc;

/* space for normal forms */

wdata = *(head+rank-1);

CALLOC(f,wdata->maxlength+OVERHEAD,weylletter_t,;)
CALLOC(f1,wdata->maxlength+OVERHEAD,weylletter_t,;)

CALLOC(v,rank,weylpiece_t,;)
CALLOC(v1,rank,weylpiece_t,;)
CALLOC(v2,rank,weylpiece_t,;)

lc = (quotsize_t *)(*head+1);
*lc = 1;
*(lc+1) = 1;

for (l=2;l<=rank;l++) /* for l=1 there is nothing to do */
	{
	
	wdata = *(head+l-1);
	state0 = *(table+l-1);
	
	state = state0 + l+1;
	newcell = state0 + 2*(l+1);
	firstfree = 2;
	
	lc = (quotsize_t *)(wdata+1);
	*lc = 1;
	lc++;
	
	/* finish the second cell */
	
	for (s=1;s<l;s++)
		{
		if (Cox(s,l)==2)	/* remain in the same state */
			{
			state[s] = -s;
			}
		else				/* construct new cell */
			{
			state[s] = firstfree++;
			*newcell = s;
			newcell[s] = 1;
			newcell += l+1;
			newlength++;
			}
		}
	
	/* fill in other cells */
	
	*lc = 1;
	lc++;
	thislength = newlength;
	newlength = 0;
	xc = 0;
	
	for (x=2;x<(*wdata).quotsize;x++)
		{
		
		state += l+1;
		
		/* put normal form of x in f, and length in p */
		
		*f = '\0';
		p = NbrToStringpiece(l,x,f);
		
		/* deprive x of its first term */
		
		RStringToArray(f+p-1,v);	
		
		for (s=1;s<=l;s++)
			{
			if (state[s]!=0)
				continue;	/* has already been filled in */
			Prod(v,s,v1);
			if (Lastterm(v1) != s)	/* v1 is not in normal form */
				{
				ArrayToString(v1,f1+1);
				*f1 = l;
				StringToArray(f1,v1);
				if (Firstterm(v1) == l) 	/* the new v1 is in X */
					{
					state[s] = v1[l-1];
					cell = state0 + (l+1)*v1[l-1];
					cell[s] = x;
					}
				else		/* remain in the same coset */
					state[s] = -Firstterm(v1);
				continue;
				}
			Prod(v1,*(f+1),v2);
			if (NaturalOrder(v2,v1))	/* v2 is not reduced */
				{
				ArrayToString(v2,f1+1);
				*f1 = l;
				StringToArray(f1,v2);
				if (Firstterm(v2) == l) 	/* create new cell */
					{
					state[s] = firstfree++;
					*newcell = s;
					newcell[s] = x;
					newcell += l+1;
					newlength++;
					}
				else		/* remain in the same coset */
					state[s] = -Firstterm(v2);
				}
			else			/* check for rank 2 */
				{
				*f = s;
				t = f[p-1];	/* second term in the normal form */
				for (j=2;j<=p;j++)
					if (f[j] != f[j-2])	/* not in rank 2 */
						{
						state[s] = firstfree++;
						*newcell = s;
						newcell[s] = x;
						newcell += l+1;
						newlength++;
						goto nexts;
						}
						
				/* at this point we are in rank 2 */
				
				if (p+1 < Cox(t,l))	/* make new cell */
					{
					state[s] = firstfree++;
					*newcell = s;
					newcell[s] = x;
					newcell += l+1;
					newlength++;
					}
				else
					state[s] = -t;
				}
				
			nexts :
				continue;
				
			}	/* next s */
		
		if (thislength == ++xc)
			{
			*lc = thislength;
			lc++;
			thislength = newlength;
			newlength = 0;
			xc = 0;
			}
			
		}	/* next x */
	
	}	/* next l */

free((char *)f);
free((char *)f1);

} /* end of MakeAutomaton */


int ArrayToString(w,f)

/*********************************************************************

Given a weylpiece_t array w and a string f, this function writes the 
normal form of w in f, and returns the length of the corresponding 
Weyl group element. The string is terminated with a '\0' so that it can
be treated as a standard unix string.

*********************************************************************/

	weylarr_t w;
	weylword_t f;

{
int l,p=0;

*f = '\0';

for (l=rank;l>0;l--)
	p += NbrToStringpiece(l,w[l-1],f+p);
	
/* reverse f */

{
register int i;
register weylletter_t c;

for (i=0;i<=p/2;i++)
	{
	c=f[i];
	f[i] = f[p-i];
	f[p-i] = c;
	}
}
	
return(p);

}	/*end of ArrayToString */


int NbrToStringpiece(l,x,f)

/***********************************************************************

	WARNING : returns string right-to-left !!

This function returns in the string pointed by f+1 the normal form of the
element x in the l-th subquotient. The normal form is written right-to-
left, since it is found that way, and also for ease of updating. The
value returned by the function is the length of the normal form.

***********************************************************************/

	rankid_t l;
	weylpiece_t x;
	weylword_t f;
	
{
state_t *state;
int p;

for (p=0;x != 0;p++)
	{
	state = *(table+l-1) + x*(l+1);
	*(++f) = *state;
	x = *(state + *state);
	}
	
return(p);

}	/* end of NbrToStringpiece */


Shift(w,s)

	weylarr_t w;
	generator_t s;

/******************************************************************

	Replaces w by ws. Left shifts are less efficient (the best way is 
to use the inverses, once MakeInverse has been called). Return value is 
0 if the length goes down, 1 if the length goes up.

	NOTE : should perhaps be modified to return -1 if length goes down

*******************************************************************/

{
state_t *state;
register rankid_t j;

for (j=rank-1;j>=0;j--)
	{
	state = *(table+j) + w[j]*(j+2) + s;
	if (*(state) >= 0)
		{
		if (*state < w[j])
			{
			w[j] = *state;
			return(0);
			}
		else
			{
			w[j] = *state;
			return(1);
			}
		}
	else
		{
		s = -*(state);
		}
	}

}	/* end of Shift */


StringShift(w,f)

/*******************************************************************

	This function shifts w by a whole string f.
	
	NOTE : some return value should be defined !

********************************************************************/

	weylarr_t w;
	weylword_t f;

{

generator_t s;

while ((s = *f++)!='\0')
	{
	Shift(w,s);
	}

}	/* end of StringShift */


StringToArray(f,w)

/*****************************************************************

	Given a character string (terminated by '\0'), not necessarily
a normal form, this function returns in w the array representing the
corresponding element of the Weyl group. 

******************************************************************/

	weylword_t f;
	weylarr_t w;

{
generator_t s;
rankid_t l;

for (l=0;l<rank;l++)
	w[l] = 0;
	
while ((s= *f++) != '\0')
	{
	Shift(w,s);
	}

}	/* end of StringToArray */


RStringToArray(f,w)

/*****************************************************************

	Given a character string (initiated by '\0'), not necessarily
a normal form, this function returns in w the array representing the
corresponding element of the Weyl group. 

	WARNING : f is assumed to point to the END of the string

******************************************************************/

	weylword_t f;
	weylarr_t w;

{
generator_t s;
rankid_t l;

for (l=0;l<rank;l++)
	w[l] = 0;
	
while ((s= *f--) != '\0')
	{
	Shift(w,s);
	}

}	/* end of RStringToArray */


Prod(w,s,ws)

/******************************************************************

	This function returns in ws the product of the element w and the
generator s. It returns 1 if the length goes up, 0 if the length
goes down.

*******************************************************************/

	weylarr_t w,ws;
	generator_t s;

{
rankid_t l;

for (l=1;l<=rank;l++)
	ws[l-1] = w[l-1];

return(Shift(ws,s));

}	/* end of Prod */


Inverse(w,wi)

/*****************************************************************

	Returns in wi the array corresponding to the inverse of w. It
	uses invstring (assumed to have been allocated) for 
	intermediate storage of the normal form.

*****************************************************************/

	weylarr_t w,wi;

{
int p;

p = ArrayToString(w,invstring+1); 

/* The first term of invstring is used as null-terminator */

RStringToArray(invstring+p,wi);

return(0);

}	/* end of Inverse */



Mult(v,w,vw)

/*****************************************************************

	Returns in vw the array corresponding to the product of the
	elements v and w. Uses multstring for intermediate storage.

******************************************************************/

	weylarr_t v,w,vw;

{

rankid_t l;

ArrayToString(w,multstring);

for (l=0;l<rank;l++)
	vw[l] = v[l];
	
return(StringShift(vw,multstring));

}	/* end of Mult */


Power(a,p,ap)

/*
  Computes the p-th power of a into ap.
*/

     weylarr_t a,ap;
     unsigned long p;

{

  rankid_t l;
  unsigned long hi_bit,q;

  if (p == 0)
    {
      for (l=0;l<rank;l++)
	{
	  ap[l] = 0;
	}
      return(0);
    }

  q = p;
  hi_bit = 0;
  hi_bit = ~(~hi_bit >> 1);  /* set the highest bit; size-independent ! */

  for (; (~p & hi_bit); p <<= 1)  /* shift p up to high bits */
    ;

  for (l=0;l<rank;l++)  /* put a in power_arr */
    {
      power_arr[l] = a[l];
    }

  for (q >>= 1; q; q >>= 1) 
    {
      p <<= 1;
      Mult(power_arr,power_arr,power_arr);  /* ap -> ap^2 */
      if (p & hi_bit)
	{
	  Mult(power_arr,a,power_arr);
	}
    }

  for (l=0; l<rank; l++)  /* put answer in ap */
    {
      ap[l] = power_arr[l];
    }

  return(0);

}   /* end of Power */


int StringMult(f,g,fg)

/***********************************************************************

	Multiplies the strings f,g into fg. Returns the length of fg.

************************************************************************/

	weylword_t f,g,fg;

{

StringToArray(f,fgarray);
StringShift(fgarray,g);
return(ArrayToString(fgarray,fg));

}	/* end of StringMult */



NormalForm(f,nf)

/********************************************************************

	Writes the normal form of f into nf. Returns the length of the 
normal form.

*********************************************************************/

	weylword_t f,nf;

{

StringToArray(f,nfarray);
return (ArrayToString(nfarray,nf));

}	/* end of NormalForm */



generator_t Firstterm(w)

/******************************************************************

	Returns the first term in the normal form of w.

*******************************************************************/

	weylarr_t w;

{
rankid_t j;

for (j=0;j<rank;j++)
	{
	if (w[j] == 0)
		continue;
	return(j+1);
	}

return(0);

}	/* end of Firstterm */



generator_t Lastterm(w)

/******************************************************************

	Returns the last term in the normal form of w.

*******************************************************************/

	weylarr_t w;

{
rankid_t l;

for (l=rank;l>0;l--)

	{
	if (w[l-1] == 0)
		continue;
	return(*(*(table+l-1) + w[l-1]*(l+1)));
	}

return(0);
	
}	/* end of Lastterm */


Longest(f)

/**********************************************************************

	Writes in f the longest element of W.

***********************************************************************/

	weylword_t f;

{

rankid_t l;
state_t *state;
weylpiece_t x;

f += head[rank-1]->maxlength;
*(f) = 0;

for (l=rank;l>0;l--)
	{
	for (x=head[l-1]->quotsize-1;x!=0;)
		{
		state = *(table+l-1) + x*(l+1);
		*(--f) = *state;	/* last term */
		x = *(state + *state);
		}
	}
	
return(head[rank-1]->maxlength);

}	/* end of Longest */



NaturalOrder(v,w)

/*********************************************************************

	Returns 1 if the integer (virtually) associated to v is smaller
	or equal to the corresponding integer for w, 0 otherwise. This 
	amounts to inverse lexicographical order on the arrays.

**********************************************************************/

	weylarr_t v,w;

{
rankid_t j;

for (j=rank-1;(j>=0 && v[j]== w[j]);j--)
	;

if (j == -1)
	return(1);

return(v[j]<w[j] ? 1 : 0);

} /* end of NaturalOrder */


MakeBetti()

/*
  Fills in the numbers in betti[] and qbetti[]. Although the necessary data
  could be gotten from BettiData(), we have used head[].
*/

{
  lengthcount_t j,m;
  rankid_t l;
  quotsize_t *q;
  weylsize_t *bn;

  bn = qbetti[rank-1];
  m = head[rank-1]->maxqlength;
  *(bn++) = m;
  q = (quotsize_t *)(head[rank-1]+1);

  for (j=0; j<=m; j++)
    {
      *(bn++) = (weylsize_t)(*(q++));
    }

  for (l=rank-1; l>0; l--)
    {
      bn = qbetti[l-1];

      /* put subquotient in qbetti[l-1] */
    
      m = head[l-1]->maxqlength;
      *(bn++) = m;
      q = (quotsize_t *)(head[l-1]+1);

      for (j=0; j<=m; j++)
	{
	  *(bn++) = (weylsize_t)(*(q++));
	}

      /* multiply with previous one */

      BPolMult(qbetti[l-1],qbetti[l],qbetti[l-1]);
    }

  for (j=0; j<=maxlength; j++)
    {
      betti[j] = qbetti[0][j+1];
    }

}				/* end of MakeBetti */


InitAutomaton()

/***************************************************************** 
   
    This function allocates and initializes memory space for
    the construction of the finite state automata which will
    carry out the Weyl group computations. The transition tables 
    will be constructed by MakeAutomaton.

    InitializeAutomaton allocates space for each of the pointers
    in the two pointer-arrays pointed by head and table, each
    consisting of rank pointers. Each head points to a structure
    of type WeylData, followed by maxqlength+1 quotsize_t,
    which will hold the numbers of elements of each length in the
    subquotient. Each table points to an array of (l+1)*quotsize
    quotsize_t, the transition tables.

******************************************************************/

{

initdata_t idata;
weyldata_t *wdata,wdata1;
rankid_t l;
int j;
quotsize_t *lc;
state_t *state;
char x;
char *ptr;

CALLOC(head,rank,weyldata_t *,;)
CALLOC(table,rank,state_t *,;)
x=type;

for (l=rank;l>0;l--)
    {

    /*allocate memory space and fill in partially*/
    
    GetInitData(x,l,&idata);
    CALLOC(ptr,sizeof(weyldata_t)+(idata.maxqlength+1)*sizeof(quotsize_t),char,;)
    wdata = (head[l-1] = (weyldata_t *)ptr);
    wdata->quotsize = idata.quotsize;
    wdata->maxqlength = idata.maxqlength;
    lc = (quotsize_t *)(wdata+1);
    for (j=0;j<=idata.maxqlength;j++)
        *(lc+j)=0;
    x = idata.subtype;
    }

wdata1.maxlength = 0;

for (l=1;l<=rank;l++)
    {
    
    /*finish headers*/
    
    wdata = head[l-1];
    wdata->maxlength = wdata1.maxlength + wdata->maxqlength;
    wdata1 = *wdata;
    }

for (l=1;l<=rank;l++)
    {
    
    /*allocate memory and initialize automata*/
    
    wdata = head[l-1];
    
    CALLOC(table[l-1],(l+1)*wdata->quotsize,state_t,;)
    state = table[l-1];
    /* returns a block initialized to zero */
    
    for (j=1;j<l;j++) /* write first cell */
    	{
    	state[j] = -j;
    	}
    state[l] = 1;
    state[l+1] = l; /* first term of second cell */
    }

wdata = head[rank-1];

return(0);

} /* end of InitAutomaton */


DestroyAutomaton()

/*******************************************************************

	This function frees the memory space allocated by 
	InitAutomaton.

********************************************************************/

{

rankid_t j;

for (j=0;j<rank;j++)
	{
	free(head[j]);
	free(table[j]);
	}

free(head);
free(table);

return(0);

}	/* end of DestroyAutomaton */


GetInitData(x,l,idata)

/******************************************************************

Returns in idata a pointer to a structure of type InitData for 
the given type and rank. For simplicity we assume type and rank 
have already been filtered so that input will always fall into 
one of the given cases.

******************************************************************/

    typeid_t x;
    rankid_t l;
    initdata_t *idata;

{
  idata->quotsize = GetQuotSize(x,l);
  idata->maxqlength = GetMaxql(x,l);
  idata->subtype = GetSubType(x,l);
  return(0);
}	/* end of GetInitData */


Cox(i,j)

/****************************************************************** 

Returns the (i,j)-entry of the Coxeter matrix 

******************************************************************/

	int i,j;

{
if (i >= j)
	return((i == j) ? 1 : Cox(j,i));

/* from now on we have i<j */

switch(type){

   case 'A':
       return((j-i == 1) ? 3 : 2);
   case 'B':
       return((j-i == 1) ? ((i == 1) ? 4 : 3) : 2);
   case 'C':
       return((j-i == 1) ? ((i == 1) ? 4 : 3) : 2);
   case 'D':
       return((j-i == 1) ? ((i == 1) ? 2 : 3) : 
	        (j-i == 2) ? ((i == 1) ? 3 : 2) : 2);
   case 'E':
       return((j-i == 1) ? ((i <= 2) ? 2 : 3) : 
	        (j-i == 2) ? ((i <= 2) ? 3 : 2) : 2);
   case 'F':
       return((j-i == 1) ? ((i == 2) ? 4 : 3) : 2);
   case 'G':
       return(6);
   case 'H':
       return((j-i == 1) ? ((i == 1) ? 5 : 3) : 2);
        }
}	/* end of Cox */


typeid_t CompType(supp)

/*
  Given a subset of S, flagged by supp, assumed to be CONNECTED, this
  function returns the type of the corresonding connected component.
*/

     sflags_t supp;

{
switch(type)
  {
  case 'A':			/* subset is always of type A */
    return('A');
    break;
  case 'B':case'C':		/* subset is of type A unless it contains 1 */
    return((~supp)&1 ? 'A' :
	   (supp == 1) ? 'A' : 'B');
    break;
  case 'D':
    /* subset is of type A unless it contains both 1 and 2 */
    return((~supp)&3 ? 'A' : 'D');
    break;
  case 'E':
    /* if subset contains 1 and 2, it is of type E; 2 and 3 but not 1,
       of type D; otherwise, of type A */
    return((~supp)&3 ? ((~supp)&6 ? 'A' : 'D') : 'E');
      break;
  case 'F':
    switch(supp)
      {
      case 7:case 15:
	return('F');
	break;
      case 6:case 14:
	return('B');
	break;
      default:
	return('A');
	break;
      }
    break;
  case 'G':
    switch(supp)
      {
      case 3:
	return('G');
	break;
      default:
	return('A');
	break;
      }
    break;
  case 'H':
    return((~supp)&1 ? 'A' :
	   (supp == 1) ? 'A' : 'H');
    break;
  }

  return(0);

}    /* end of CompType */


sflags_t Component(f,s)

/*
  Returns the connected component of s in f. It uses the array star, which
  for each s contains the star of s in S
*/

     sflags_t f;
     generator_t s;

{
  generator_t t;
  sflags_t c,tf;

  c = 1;
  c <<= s-1;

  if (c&~f)    /* s is not in f */
    return(0);

  for(t=1,tf=1;t<=rank;)
    {
      if ((tf&f&~c)&&(c&star[t-1]))
	{
	  c |= tf;
	  tf = 1;
	  t=1;
	}
      else
	{
	  t++;
	  tf <<= 1;
	}
    }

  return(c);
}    /* end of Component */


typeid_t GetSubType(x,l)

/*
  Returns the type of the first subgroup of the group of type x,
  rank l
*/

{
  switch (x)
    {
    case 'A':  /* rank is >= 0 */
      return(l==0 ? '*' : 'A');
      break;
    case 'B':  /* rank is >= 2 */
      return(l==2 ? 'A' : 'B');
      break;
    case 'C':  /* rank is >= 2 */
      return(l==2 ? 'A' : 'C');
      break;
    case 'D':  /* rank is >= 2 */
      return(l==2 ? 'A' : 'D');
      break;
    case 'E':  /* rank is between 3 and 8 */
      return(l==3 ? 'D' : 'E');
      break;
    case 'F':  /* rank is 3 or 4 */
      return(l==3 ? 'A' : 'F');
      break;
    case 'G':  /* rank is 2 */
      return('A');
      break;
    case 'H':  /* rank is 2, 3 or 4 */
      return(l==2 ? 'A' : 'H');
      break;
    }

  return(0);

}    /* end of GetSubType */


quotsize_t GetQuotSize(x,l)

/*
  Returns the size of the last subquotient of the group of
  type x, rank l
*/

{
  switch (x)
    {
    case 'A':
      return((quotsize_t)(l+1));
      break;
    case 'B':
      return((quotsize_t)(2*l));
      break;
    case 'C':
      return((quotsize_t)(2*l));
      break;
    case 'D':
      return((quotsize_t)(l==2 ? 2 : 2*l));
      break;
    case 'E':
      switch(l)
	{
	case 3:
	  return(3);
	  break;
	case 4:
	  return(10);
	  break;
	case 5:
	  return(16);
	  break;
	case 6:
	  return(27);
	  break;
	case 7:
	  return(56);
	  break;
	case 8:
	  return(240);
	  break;
	}
      break;
    case 'F':
      switch(l)
	{
	case 3:
	  return(8);
	  break;
	case 4:
	  return(24);
	  break;
	}
      break;
    case 'G':
      return(6);
      break;
    case 'H':
      switch(l)
	{
	case 2:
	  return(5);
	  break;
	case 3:
	  return(12);
	  break;
	case 4:
	  return(120);
	  break;
	}
      break;
    }

  return(0);
  
}    /* end of GetQuotSize */


lengthcount_t GetMaxql(x,l)

/*
  Returns the maximum length in the last subquotient of the group of
  type x, rank l
*/

     typeid_t x;
     rankid_t l;

{
  switch (x)
    {
    case 'A':
      return((lengthcount_t)l);
      break;
    case 'B':
      return((lengthcount_t)(2*l-1));
      break;
    case 'C':
      return((lengthcount_t)(2*l-1));
      break;
    case 'D':
      return(l == 2 ? 1 : (lengthcount_t)(2*l-2));
      break;
    case 'E':
      switch(l)
	{
	case 3:
	  return(2);
	  break;
	case 4:
	  return(6);
	  break;
	case 5:
	  return(10);
	  break;
	case 6:
	  return(16);
	  break;
	case 7:
	  return(27);
	  break;
	case 8:
	  return(57);
	  break;
	}
      break;
    case 'F':
      switch(l)
	{
	case 3:
	  return(6);
	  break;
	case 4:
	  return(15);
	  break;
	}
      break;
    case 'G':
      return(5);
      break;
    case 'H':
      switch(l)
	{
	case 2:
	  return(4);
	  break;
	case 3:
	  return(10);
	  break;
	case 4:
	  return(45);
	  break;
	}
      break;
    }

return(0);
  
}    /* end of GetMaxql */


BettiData(x,l,b)

/*
  Returns in the array b the betti numbers of the last subquotient of the
  group of type x, rank l.
*/

     typeid_t x;
     rankid_t l;
     bettipol_t b;

{
  lengthcount_t m,j;

  m = GetMaxql(x,l);
  
  for (j=0;j<=m;j++)  /* the betti number is at least 1 */
    {
      b[j] = 1;
    }

  /* correct betti numbers if necessary */
  
  switch (x)
    {
    case 'A':  /* no correction */
      break;
    case 'B':  /* no correction */
      break;
    case 'C':  /* no correction */
      break;
    case 'D':  /* b[l-1] = 2 */
      b[l-1]++;
      break;
    case 'E':
      switch (l)
	{
	case 3:
	  break;
	case 4:
	  for (j=2;j<=4;j++)
	    b[j]++;
	  break;
	case 5:
	  for (j=3;j<=7;j++)
	    b[j]++;
	  break;
	case 6:
	  for (j=4;j<=12;j++)
	    b[j]++;
	  b[8]++;
	  break;
	case 7:
	  for (j=5;j<=22;j++)
	    b[j]++;
	  for (j=9;j<=18;j++)
	    b[j]++;
	  break;
	case 8:
	  for (j=6;j<=51;j++)
	    b[j]++;
	  for (j=10;j<=47;j++)
	    b[j]++;
	  for (j=12;j<=45;j++)
	    b[j]++;
	  for (j=16;j<=41;j++)
	    b[j]++;
	  for (j=18;j<=39;j++)
	    b[j]++;
	  for (j=22;j<=35;j++)
	    b[j]++;
	  b[28]++;
	  b[29]++;
	  break;
	}
      break;
    case 'F':
      switch (l)
	{
	case 3:
	  b[3]++;
	  break;
	case 4:
	  for (j=4;j<=11;j++)
	    b[j]++;
	  break;
	}
      break;
    case 'G':
      break;
    case 'H':
      switch (l)
	{
	case 2:
	  break;
	case 3:
	  b[5]++;
	  break;
	case 4:
	  for (j=6;j<=39;j++)
	    b[j]++;
	  for (j=10;j<=35;j++)
	    b[j]++;
	  for (j=16;j<=29;j++)
	    b[j]++;
	  break;
	}
      break;
    }
}				/* end of BettiData */


BettiPol(f,p)

/*
  Finds the bettipol_t corresponding to the subgroup flagged by f. The result
  is returned in p. It uses the arrays bpol1_work and bpol2_work.
*/

     sflags_t f;
     bettipol_t p;

{
  sflags_t c,fs;
  generator_t s;
  typeid_t x;
  rankid_t l;

  if (f == 0)
    {
      *p = 0;
      p[1] = 1;
      return(0);
    }

  /* put largest element of f in s */

  for (fs=f,s=0;fs;fs>>=1,s++)
    ;

  c = Component(f,s);

  if (c!=f)  /* conclude by recursion */
    {
      BettiPol(f&~c,p);
      BettiPol(c,bpol1_work);
      BPolMult(p,bpol1_work,p);
      return(0);
    }

  /* now f is connected */

  fs = 1 << s-1;
  
  BettiPol(f&~fs,p);
  x = CompType(f);
  l = BitCount((lflags_t)f);
  *bpol2_work = GetMaxql(x,l);
  BettiData(x,l,bpol2_work+1);
  BPolMult(p,bpol2_work,p);
  return(0);

  /* now */

}    /* end of BettiPol */


BPolAdd(p,q,sum)

/*
  Adds two bettipol_t p and q; the result is returned in sum. This
  function does not need a workstring, since the values of p and q
  are used only once, so they may be overwritten.
*/

     bettipol_t p,q,sum;

{
  register int j;
  int d;

  if (*p == *q)  /* degree is hard to get */
    {
      for (d = *p;(p[d+1]+q[d+1] == 0) && (d>0);d--)
	;
    }
  else  /* degree is biggest of two */
    {
      if (*p < *q)
	{
	  d = *q;
	}
      else
	{
	  d = *p;
	}
    }

  *sum = d;

  for (j=1;j<=d+1;j++)
    {
      sum[j] = p[j] + q[j];
    }
}    /* end of BPolAdd */


BPolMult(p,q,pq)

/*
  Multiplies two bettipol_t p and q; the result is returned in pq
*/

     bettipol_t p,q,pq;

{
  register weylsize_t k;
  register int i,j;
  
  *multwork = *p + *q;  /* degrees add up */

  for (j=0;j<=*multwork;)  /* clean up pq in the appropriate range */
    {
      multwork[++j] = 0;
    }

  for (j=0;j<=*p;)
    {
      k = p[++j];
      for (i=0;i<=*q;i++)
	{
	  multwork[i+j] += k*q[i+1];
	}
    }

  for (j=0;j<=*multwork+1;j++)
    {
      pq[j] = multwork[j];
    }
  
}    /* end of BPolMult */


BPolDiv(num,den,quot)

/*
  Divides num by den; it is assumed that the quotient is exact; it uses
  the string divwork.
*/

     bettipol_t num,den,quot;

{
  register weylnbr_t k;
  register int i,j;

  for (j=0;j<=*num+1;j++)
    divwork[j] = num[j];

  *quot = *num - *den;
  
  for (j = *quot;j >= 0;j--)
    {
      k = divwork[*divwork+1]/den[*den+1];
      (*divwork)--;
      for (i=0;i<=*den;i++)
	{
	  divwork[i+j+1] -= k*den[i+1];
	}
      quot[j+1] = k;
    }

}    /* end of BPolDiv */


BitCount(f)
     
     /**********************************************************************
       
       Returns the number of set bits in f.
       
       ***********************************************************************/
     
     lflags_t f;
     
{
  register int count;
  
  for (count=0;f!=0;count++)
    {
      f = f & (f-1);	/* see K&R */
    }
  
  return(count);
}	/* end of BitCount */


WeylnbrToArray(w,a)

/*
  This function computes the normal form of an element directly from its
  weylnbr. It is intended to allow for shorthand notation in input.

  First, the length of w is determined using betti[]. Then, the idea is to
  describe the lexicographical order in terms of the natural tree structure
  on W.

  NOTE : in this function and the next one, we use weylsize_t instead of
         weylnbr_t, because weylnbr_t may be smaller; see comments in
	 types.h
*/

     weylsize_t w;
     weylarr_t a;

{

  lengthcount_t p,q;
  weylsize_t m;
  rankid_t l;
  quotsize_t k;

  /* find length of w in p */

  p = 0;
  m = betti[0];

  while (w >= m)
    {
      w -= m;
      p++;
      m = betti[p];
    }

  /* now w holds the adress of the element starting from first in length p */

  for (l=0;l<rank-1;l++)  /* find corresponding piece of w */
    {
      for (k=0;w<m;k++)
	{
	  q = p - xlength[l][k];
	  if ((q>=0)&&(q<=*qbetti[l+1]))
	    {
	      m -= qbetti[l+1][q+1];
	    }
	}
      a[l] = xlist[l][k-1];  /* piece was found */
      p -= xlength[l][k-1];  /* length of remaining piece */
      w -= m;           /* put relative adress in w */
      m = qbetti[l+1][q+1];  /* new upper bound */
    }

  /* Find last piece of w (w holds adress of this starting from first
     element in length p) */

  a[rank-1] = 0;

  for (q=0; q<p; q++)
    {
      a[rank-1] += qbetti[rank-1][q+1];
    }

  a[rank-1] += w;

}    /* end of WeylnbrToArray */


ArrayToWeylnbr(a,w)

/*
  Gives the lexicographical order number of a weylarr_t.
*/

     weylarr_t a;
     weylsize_t *w;

{

  rankid_t l;
  lengthcount_t p,q;
  quotsize_t k,*k1;
  weylpiece_t x;

  /* find length of w */

  p = 0;

  for (l=0; l<rank; l++)
    {
      k1 = (quotsize_t *)(head[l]+1);

      x = a[l];
      q = 0;
      k = k1[0];

      while (x >= k)
	{
	  x -= k;
	  k = k1[++q];
	}

      p += q;
    }

  /* find relative adress of w */

  *w = firstlength[p+1];

  for (l=0; l<rank-1; l++)
    {
      x = a[l];
      for (k=0; x != xlist[l][k]; k++)
	{
	  q = p - xlength[l][k];
	  if ((q >= 0)&&(q <= *qbetti[l+1]))
	    {
	      *w -= qbetti[l+1][q+1];
	    }
	}
      p -= xlength[l][k];  /* length of remaining piece */
    }

  /* now w holds the upper bound for the elements with same beginning and same
     length as ours */

  x = a[rank-1];

  for (q = 0; q < p; q++)
    {
      x -= qbetti[rank-1][q+1];
    }

  *w -= (qbetti[rank-1][p+1] - x);

  return(0);

}    /* end of ArrayToWeylnbr */
